
#include "regs.h"
#include "apu.h"

u8 apuwaitfor;
u8 apukickval;

/* SPC700 boot ROM protocol */

u16 isapuinited(void) {
	return APUI00 == 0xAA && APUI01 == 0xBB;
}

void apuwaitinit(void) {
	// inline // while(!isapuinited()) /*busyloop*/ ;
	while(APUI00 != 0xAA) /*busyloop*/ ;
	while(APUI01 != 0xBB) /*busyloop*/ ;

	apuwaitfor = 0xAA;
	apukickval = 0xCC;
}

void apuwait(void) {
	while(APUI00 != apuwaitfor) /*busyloop*/ ;
}

void apukick(u16 addr, u8 mode) {
	apuwait();

	APUI02 = addr & 255;
	APUI03 = addr >> 8;
	APUI01 = mode;
	APUI00 = apukickval;

	apuwaitfor = apukickval;
}

void apuloadstart(u16 addr) {
	apukick(addr, !!1);

	apukickval = 0;
}

void apusend(u8 byte) {
	apuwait();

	APUI01 = byte;
	APUI00 = apukickval;

	apuwaitfor = apukickval;
	++apukickval;
}

void apuend(void) {
	apukickval = (apukickval + 1) | 1; // larger (as if >8bit) than to-be value but not zero
}

void apuexec(u16 addr) {
	apukick(addr, !!0);

	// advance apukickval for trildr
	apukickval = (apukickval + 2) | 1; // larger (as if >8bit) than to-be value but not zero
}

void apuload(u16 size, u16 dst, u8 *src) {
	apuloadstart(dst);
	while(size) {
		apusend(*src);
		++src;
		--size;
	}
	apuend();
}

void apureg(u8 index, u8 val) {
	apuloadstart(0x00F2);
	apusend(index);
	apusend(val);
	apuend();
}

/* trildr protocol */

void triload(u16 size, u16 dst, u8 *src) {
	u8 *p = src;
	u8 *ptail = p + size;

	--p;

	apuwait();
	APUI02 = dst;
	APUI03 = dst >> 8;
	APUI01 = 0x01; // non-zero
	APUI00 = apukickval;
	apuwaitfor = apukickval;
	apukickval = 0x00;

	while(p < ptail) {
		apuwait();
		APUI01 = *++p;
		APUI02 = *++p;
		APUI03 = *++p;
		APUI00 = apukickval;
		apuwaitfor = apukickval;
		apukickval += 3;
	}

	apukickval = (apukickval + 1) | 1; // if next is exec, "| 1" is not required...
}

void triexec(u16 addr) {
	apuwait();
	APUI02 = addr;
	APUI03 = addr >> 8;
	APUI01 = 0;
	APUI00 = apukickval;
	apuwaitfor = apukickval;
	apukickval = (apukickval + 2) | 1; // advance kick for next protocol...
}
