
#include "regs.h"
#include "apu.h"

//#define DEBUG

#define REP_IWAIT 20
#define REP_RWAIT 3

#define MAXBOMBS 9
#define MAXKEYS 12   // including header byte

#define ECHOSIZE (0 * 2048 ?: 0*256 ?: 0) // n*2048(n frames echo), 256(1 sample echo) or zero(no echo)
#define DIRAT (ECHOSIZE + 0x0200)
#define BRRAT (ECHOSIZE + 0x0300)

#define BI_B  0
#define BI_Y  1
#define BI_PU 4
#define BI_PD 5
#define BI_PL 6
#define BI_PR 7
#define BI_A  8
#define BI_X  9
#define BI_L  10
#define BI_R  11

#define ARRAY_SIZE(a) (sizeof(a)/sizeof((a)[0]))

extern void *nmi_handler;
extern u8 textchr[], annivchr[], titlechr[], titlenam[], descchr[], descnam[], trildr[], apuzero[], sounds[], sounds_bin[];
extern u16 sizeof_textchr, sizeof_annivchr, sizeof_titlechr, sizeof_descchr, sizeof_trildr, sizeof_sounds, sizeof_sounds_bin, sizeof_apuzero;
extern u16 sizeof_15th;
extern u8 buildstr[];

const u16 soundrates[] = {
	0x1000* 8000/32000, // 15th
	0x1000*16000/32000, // accept
	0x1000*16000/32000, // invalidate
	0x1000*10000/32000, // fall
	0x1000* 6000/32000, // gover
	0x1000*12000/32000, // goodstart
	0x1000*12000/32000, // kaboom
	0x1000* 8000/32000, // obj
	0x1000* 2000/32000, // sadstart
	0x1000* 8000/32000, // type
	0x1000*12000/32000, // succeed
};
const struct difficulty {
	u16 spawnrate;
	u8 fallspeed;
	u8 minkeys;
	u8 maxkeys;
	u8 maxpars;
	u8 bombs;
	u8 nkeycands;
	u8 keycands[12];
} difficulties[] = {
	// rate spd min max pr norm ncand cands...
	{0x0100,  5,  3,  3, 3,  10,  2, {BI_B, BI_A}},
	{0x0100,  5,  3,  3, 3,  10,  4, {BI_PU, BI_PD, BI_B, BI_A}},
	{0x0200,  5,  3,  3, 3,  15,  4, {BI_PL, BI_PR, BI_B, BI_A}},
	{0x0200,  5,  3,  5, 5,  20,  4, {BI_PU, BI_PD, BI_PL, BI_PR}},
	{0x0200,  7,  4,  4, 3,  20,  4, {BI_PU, BI_PD, BI_B, BI_A}},
	{0x0200,  7,  4,  4, 3,  20,  6, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A}},
	{0x0200,  7,  4,  5, 3,  20,  6, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A}},
	{0x0280,  5,  4,  6, 4,  20,  6, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A}},
	{0x0200,  7,  4,  5, 3,  30,  6, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A}},
	{0x0800, 16,  1,  1, 9,  30,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}}, // introduce X/Y in this easy wave
	{0x0280,  7,  4,  5, 3,  20,  4, {BI_X, BI_Y, BI_A, BI_B}},
	{0x0280,  7,  4,  6, 2,  30,  6, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A}},
	{0x0280,  7,  4,  6, 2,  20,  6, {BI_PU, BI_PD, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280,  7,  4,  6, 3,  30,  6, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A}},
	{0x0280,  7,  4,  6, 3,  20,  6, {BI_PU, BI_PD, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280,  7,  4,  5, 3,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280,  7,  4,  6, 3,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},

	{0x0280, 10,  3,  5, 2,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 10,  3,  5, 3,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 10,  4,  6, 3,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 10,  5,  6, 3,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 10,  5,  6, 4,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 13,  3,  5, 2,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 13,  3,  5, 3,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 13,  4,  6, 3,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 13,  5,  6, 3,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 13,  5,  6, 4,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 13,  5,  8, 3,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 13,  5,  8, 4,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 13,  7,  8, 4,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 13,  8, 10, 4,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 13,  8, 11, 4,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 13,  9, 11, 4,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0280, 13, 10, 11, 4,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0300, 13, 10, 11, 4,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0300, 13, 10, 11, 6,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0300, 13, 10, 11, 7,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},
	{0x0300, 13, 10, 11, 8,  20,  8, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_X, BI_Y}},

	{0x1000, 16, 11, 11, 9, 300, 10, {BI_PU, BI_PD, BI_PL, BI_PR, BI_B, BI_A, BI_Y, BI_X, BI_L, BI_R}},
};

u16 vblanked;
u16 btn, pbtn, dbtn;
u16 krep_cnt, krep_last;
u16 oambuf[4/2*128+128/4/2];
u16 sound;
u16 chan;
u16 randseed;
void (*scene_think_input)(void);
void (*scene_ppu)(void);

// title/desc/prewave/gameover states
u16 waitcnt;

// title states
u16 subtix, subtidx;
u8 hdmabuf[128];
u16 secreti;

// desc states
u16 descrl;

// game states
u16 life;
u16 waveno;

// wave params
struct difficulty difficulty; // cache of difficulty
// wave states
u16 bombx[MAXBOMBS], bomby[MAXBOMBS]; // note: bomby is u12.4
u8 diskeys[MAXBOMBS*MAXKEYS];
u16 nospawnperiod;
u16 pause;
u16 kaboomx, scrscli;
s16 scrsclx, scrscly;

void vblank(void) {
	vblanked = 1;
}

// http://www.retroprogramming.com/2017/07/xorshift-pseudorandom-numbers-in-z80.html
// non-optimized way
u16 rand(void) {
	randseed ^= randseed << 7;
	randseed ^= randseed >> 9;
	randseed ^= randseed << 8;
	return randseed;
}

void dma0(u16 size, void *aaddr, u8 baddr, u8 mode) {
	DMAP0 = mode;
	BBAD0 = baddr;
	A1T0 = (u16)aaddr;
	A1B0 = ((u8*)&aaddr)[2]; // bank FIXME 816-tcc can't lsr long(32bit) correctly. void* holds bank at [2]. COMPILER-DEPENDENT!!
	DAS0 = size;
	MDMAEN = 0x01; // go DMA0!
}

// load chr from rodata
void loadchr(void) {
	VMADD = 0x0000; // vramaddr=0KW
	dma0(sizeof_textchr, textchr, 0x18, 0x01); // (u16*)0x2118=vramio, mem2io,direct,a=inc,word
	// anniv will be overwritten.
	VMADD = 0x0C00; // vramaddr=3KW
	dma0(sizeof_annivchr, annivchr, 0x18, 0x01); // (u16*)0x2118=vramio, mem2io,direct,a=inc,word
	VMADD = 0x1400; // vramaddr=5KW
	dma0(sizeof_titlechr, titlechr, 0x18, 0x01); // (u16*)0x2118=vramio, mem2io,direct,a=inc,word
}

// clear bg0
void clrbg0(void) {
	u16 zero = 0x00;

	VMADD = 0x4000; // vramaddr=16KW
#if 1
	dma0(2 * 32 * 32, &zero, 0x18, 0x09); // (u16*)0x2118=vramio, a2b,direct,a=fixed,word
#else
	{
		u16 i;
		for(i = 0; i < 32*28; ++i) {
			VMDATA = 0x7F;
		}
	}
#endif
}

// init palette
#define RGB15(r,g,b) ((r) | ((g) << 5) | ((b) << 10))
#define COLORW(a) do{ CGDATA = (a) & 255; CGDATA = (a) >> 8; }while(0)
#define COLOR1(r,g,b) do{ COLORW(RGB15(r,g,b)); }while(0)
#define COLOR4(wr,wg,wb,xr,xg,xb,yr,yg,yb,zr,zg,zb) do{ COLOR1(wr,wg,wb); COLOR1(xr,xg,xb); COLOR1(yr,yg,yb); COLOR1(zr,zg,zb); }while(0)
void loadpal(void) {
	CGADD = 0x00; // cgramaddr=0x00 (backdrop/BG1)
	COLOR4(0,0,12, 2,2,2, 7,7,7, 31,31,31); // darkblue(backdrop), darkgray, gray, white (chars)
	COLOR4(0,0,15, 8,2,24, 13,4,28, 13,1,1); // darkblue, darkpurple, purple, brown (gamebg)
	CGADD = 0x20; // cgramaddr=0x20 (BG2)
	COLOR4(0,0,0, 2,2,2, 31,8,8, 31,20,20); // <backdrop>, blacky, red, lightred (life)
	COLOR4(0,0,0, 0,0,0, 0,0,0, 31,31,31); // <backdrop>, black, black, white (bg1chars)
	COLOR4(0,0,0, 0,0,0, 0,0,0, 31,0,0); // <backdrop>, black, black, red (bg1chars:red)

	CGADD = 0x80; // cgramaddr=0x80 (OBJ0)
	COLOR4(0,0,0, 0,0,0, 8,8,8, 16,16,16); // bomb
	CGADD = 0x90; // cgramaddr=0x90 (OBJ1)
	COLOR4(0,0,0, 8,8,8, 0,0,0, 0,0,0); // done-btns
	CGADD = 0xA0; // cgramaddr=0xA0 (OBJ2)
	COLOR4(0,0,0, 31,20,20, 0,0,0, 0,0,0); // now-btns
	CGADD = 0xB0; // cgramaddr=0xB0 (OBJ3)
	COLOR4(0,0,0, 31,31,31, 0,0,0, 0,0,0); // will-btns
	CGADD = 0xC0; // cgramaddr=0xC0 (OBJ4)
	COLOR4(0,0,0, 28,24,15, 29,20,8, 29,8,8); COLOR1(29,28,16); COLOR1(31,31,31); // kaboom
}

void clrobj(void) {
	u16 i;

	for(i = 0; i < 128; ++i) {
		oambuf[i * 2 + 0] = (240 << 8) | 128; // yx
		oambuf[i * 2 + 1] = 0; // at
	}
	for(i = 0; i < 128/4/2; ++i) {
		oambuf[128 * 2 + i] = 0x5555; // x+=256
	}

	OAMADD = 0;
	dma0(0x220, oambuf, 0x04, 0x02); // 2104=OAMDATA a2b,direct,a=inc,word-same-byte
}

void init(void) {
	INIDISP = 0x8F; // forceblank bright=16

	nmi_handler = vblank;
	NMITIMEN = 0x81; /* enable NMI, enable autojoy */
}

u16 krep(u16 btn, u16 key) {
	if(!btn) {
		krep_last = 0;
		return 0;
	}
	if(!(btn & key)) {
		return 0;
	}

	if(krep_last != key) {
		krep_last = key;
		krep_cnt = 0;
		return 1;
	}

	++krep_cnt;
	if(krep_cnt == REP_IWAIT) {
		return 1;
	}
	if(krep_cnt == REP_IWAIT + REP_RWAIT) {
		krep_cnt = REP_IWAIT;
		return 1;
	}
	return 0;
}

void apuzerocall(u16 size, u16 addr) {
	if(size == 0) {
		return;
	}

	size = -size;

	apuload(sizeof_apuzero, 0x0080, apuzero);
	apuloadstart(0x0002);
	apusend(size);
	apusend(size >> 8);
	apuend();
	apuloadstart(0x0085);
	apusend(addr);
	apusend(addr >> 8);
	apuend();
	apuexec(0x80);
	apuwaitinit();
}

void soundinit(void) {
	apuwaitinit();

	apureg(0x6C, 0xE0); // reset,mute,echoro,noise=0

	apuload(sizeof_trildr, 0xFF80, trildr);

	// load dir with relocate
	{
		u16 *p = (void*)&sounds[0];
		u16 *q = (void*)(sounds + sizeof_sounds);

		apuloadstart(DIRAT);
		while(p < q) {
			u16 v = *p + BRRAT; // base of sounds_bin
			apusend(v);
			apusend(v >> 8);
			++p;
		}
		apuend();
	}

	apuexec(0xFF80);
	// load only 15th: anniv scene load rest
	//triload(sizeof_sounds_bin, BRRAT, sounds_bin);
	triload(sizeof_15th, BRRAT, sounds_bin);
	triexec(0xFFC9);
	apuwaitinit();

	// zero echobuf
	apuzerocall(ECHOSIZE, 0x0200);

	apureg(0x6D, 0x02); // echobuf@0x0200
	apureg(0x7D, ECHOSIZE / 2048); // echosize

	apureg(0x2C, 0x00); // echoL
	apureg(0x3C, 0x00); // echoR
	apureg(0x0D, 0x00); // echoFB

	// lowpass??
	apureg(0x0F, 0xFF); // FIR0...
	apureg(0x1F, 0x08);
	apureg(0x2F, 0x15);
	apureg(0x3F, 0x22);
	apureg(0x4F, 0x22);
	apureg(0x5F, 0x15);
	apureg(0x6F, 0x08);
	apureg(0x7F, 0xFF); // ...FIR7

	apureg(0x0C, 0x7F); // masterL
	apureg(0x1C, 0x7F); // masterR

	apureg(0x2D, 0x00); // pitchmod=no-ch
	apureg(0x3D, 0x00); // noise=no-ch
	apureg(0x4D, 0x00); // echoEn(no-ch)

	apureg(0x5D, DIRAT >> 8); // dir@0A00

	for(chan = 0; chan != 0x80; chan += 0x10) {
		apureg(chan | 0x00, 0); // volL
		apureg(chan | 0x01, 0); // volR
		apureg(chan | 0x02, 0); // pitchL
		apureg(chan | 0x03, 0>>8); // pitchR
		apureg(chan | 0x04, 0); // src
		apureg(chan | 0x05, 0x80); // adsrl
		apureg(chan | 0x06, 0>>8); // adsrh
		apureg(chan | 0x07, 0); // gain
	}

	apureg(0x6C, 0x20); // noreset,nomute,echoro,noise=0

	chan = 0;
}
#define SOUND_15TH    0
#define SOUND_ACCEPT  1
#define SOUND_INVALID 2
#define SOUND_FALL    3
#define SOUND_GOVER   4
#define SOUND_GSTART  5
#define SOUND_KABOOM  6
#define SOUND_OBJCTV  7
#define SOUND_SSTART  8
#define SOUND_TYPE    9
#define SOUND_SUCCESS 10

#define ADSR(ar,dr,sl,sr) (((sl) << 13) | ((sr) << 8) | 0x80 | ((dr) << 4) | (ar))
//#define PLAY 0x007F
#define PLAY ADSR(15, 0, 7, 0)
void say(u16 idx, u16 env) {
	u16 rate;
	if(!sound) { return; }
	rate = soundrates[idx];
	apureg(chan | 0x00, 127); // volL
	apureg(chan | 0x01, 127); // volR
	apureg(chan | 0x02, rate); // pitchL
	apureg(chan | 0x03, rate>>8); // pitchR
	apureg(chan | 0x04, idx); // src
	apureg(chan | 0x05, env); // adsrl
	if(env & 0x80) {
		apureg(chan | 0x06, env>>8); // adsrh
	} else {
		apureg(chan | 0x07, env & 0x7F); // gain
	}
	apureg(0x4C, 1 << (chan >> 4));
	chan = (chan + 0x10) & 0x70;
}

void title_init(void);

void gameover_think_input(void) {
	btn = JOY1;
	randseed ^= btn;
	dbtn = (pbtn ^ btn) & btn;
	pbtn = btn;

	if(waitcnt < 120) {
		++waitcnt;
	} else if(pbtn) {
		title_init();
	}
}
void gameover_ppu(void) {
	/* nothing to do... */
}

void gameover_init(void) {
	u16 i;
	u8 *p;

	// reset the last shake
	BG1HOFS = 0; BG1HOFS = 0>>8;
	BG1VOFS = 0; BG1VOFS = 0>>8;
	BG2HOFS = 0; BG2HOFS = 0>>8;
	BG2VOFS = 0; BG2VOFS = 0>>8;

	scene_think_input = gameover_think_input;
	scene_ppu = gameover_ppu;

	INIDISP = 0x8F; // forceblank bright=16

	// clear bg0 for make black
	VMADD = 0x4400;
	clrbg0();

	CGADD = 0x00; // backdrop
	COLOR1(0,0,0); // to black

	// clear with status area
	VMADD = 0x4400;
	for(i = 0; i < 32*4; ++i) {
		VMDATA = 0x2001;
	}
	for(; i < 32*30; ++i) { // +2 extra
		VMDATA = 0x2000;
	}
	for(; i < 32*32; ++i) { // all 2 extra
		VMDATA = 0x2001;
	}

	VMADD = 0x4400 + 32*12 + 11;
	for(p = "GAME OVER"; *p; ++p) {
		VMDATA = 0x2800 | *p;
	}

	if(waveno < 2) {
		VMADD = 0x4400 + 32*16 + 7;
		for(p = "Are you serious??"; *p; ++p) {
			VMDATA = 0x2400 | *p;
		}
	} else {
		VMADD = 0x4400 + 32*16 + 6;
		for(p = "You survived "; *p; ++p) {
			VMDATA = 0x2400 | *p;
		}
		if(waveno < 11) {
			VMDATA = 0x2400 | ('0' + waveno - 1);
		} else {
			// FIXME ugh... inline itoa
			u16 divisor = 10 << 3, tenbit = 1 << 3, tenth = 0, teeny = waveno - 1;
			if(99 < teeny) { teeny = 99; }
			while(9 < teeny && tenbit) {
				if(divisor <= teeny) {
					teeny -= divisor;
					tenth |= tenbit;
				}
				divisor >>= 1, tenbit >>= 1;
			}
			VMDATA = 0x2400 | ('0' + tenth);
			VMDATA = 0x2400 | ('0' + teeny);
		}
		for(p = " WAVE"; *p; ++p) {
			VMDATA = 0x2400 | *p;
		}
		if(2 < waveno) {
			VMDATA = 0x2400 | 's';
		}
	}

	clrobj();

	say(SOUND_GOVER, PLAY);

	waitcnt = 0;
}

void updobj(void) {
	u16 i, j, obi, obs, obb, oki, oks, okb, dki;
	obi = (6+MAXBOMBS*MAXKEYS)*2;
	obs = 256+(6+MAXBOMBS*MAXKEYS)/8;
	obb = 1 << ((6+MAXBOMBS*MAXKEYS)%8*2);
	oki = 6*2;
	oks = 256+6/8;
	okb = 1 << (6%8*2);
	dki = 0;

	for(i = 0; i < 128/8; ++i) {
		oambuf[i+256] = 0x5555;
	}

	for(i = 0; i < MAXBOMBS; ++i) {
		if(bomby[i] == -1) {
			oambuf[obi+0] = (240 << 8) | 128; // hide slot
		} else {
			oambuf[obi+0] = ((bomby[i] / 16 - scrscly) << 8) | (bombx[i] - scrsclx);
			oambuf[obi+1] = 0x204C;
			oambuf[obs] = (oambuf[obs] & ~obb) | (obb << 1);
			if(!pause) {
				u16 ds = diskeys[dki] >> 4, di = diskeys[dki] & 15;
				u16 x = bombx[i] + 8 - ds * 8 / 2 - scrsclx;
				for(j = 1; j <= ds; ++j) {
					oambuf[oki+0] = ((bomby[i] / 16 - 8 - scrscly) << 8) | x;
					oambuf[oki+1] = ((j < di) ? 0x2240 : (j == di) ? 0x2450 : 0x2640) + diskeys[dki+j];
					oambuf[oks] = oambuf[oks] & ~okb;
					x += 8;
					oki += 2; okb <<= 2; if(okb == 0) { ++oks; okb = 1; }
				}
			}
			obi += 2; obb <<= 2; if(obb == 0) { ++obs; obb = 1; }
		}
		dki += MAXKEYS;
	}

	if(kaboomx != 0x8000) {
		// reuse oks/okb
		i = 0 * 2;
		oks = 256+0/2/8;
		okb = 1 << (0/2%8*2);
		for(j = 0; j < 6; ++j) {
			oambuf[oks] = (oambuf[oks] & ~okb) | (okb << 1);
			okb <<= 2; if(okb == 0) { ++oks; okb = 1; }
		}
		--i;
		oambuf[++i] = (( 192       - scrscly) << 8) | (kaboomx +  0 - scrsclx); oambuf[++i] = 0x3860;
		oambuf[++i] = (( 192       - scrscly) << 8) | (kaboomx + 16 - scrsclx); oambuf[++i] = 0x3862;
		oambuf[++i] = (( 192       - scrscly) << 8) | (kaboomx + 32 - scrsclx); oambuf[++i] = 0x3864;
		oambuf[++i] = (((192 + 16) - scrscly) << 8) | (kaboomx +  0 - scrsclx); oambuf[++i] = 0x3866;
		oambuf[++i] = (((192 + 16) - scrscly) << 8) | (kaboomx + 16 - scrsclx); oambuf[++i] = 0x3868;
		oambuf[++i] = (((192 + 16) - scrscly) << 8) | (kaboomx + 32 - scrsclx); oambuf[++i] = 0x386A;
	}
}

void prewave_init(void);

u16 shake(void) {
	u16 i, j, upd = 0;

	++scrscli;
	if(scrscli < 8*2) { i = 1; j = 15; }
	else if(scrscli < 4*4 + 8*2) { i = 3; j = 7; }
	else if(scrscli < 2*16 + 4*4 + 8*2) { i = 15; j = 3; }
	else if(scrscli < 4*31 + 2*16 + 4*4 + 8*2) { i = 31; j = 3; }
	else { i = 255; j = 0; }
	if(!(scrscli & i)) {
		scrsclx = rand() & j - j / 2;
		scrscly = rand() & j - j / 2;
		++upd;
	}
	if(150 < scrscli) {
		--life;
		if(life != 0) {
			prewave_init();
		} else {
			gameover_init();
		}
	}
	return upd;
}

// FIXME ugly return value...
u16 fall(void) {
	u16 i, j, obi, oki, dki, upd = 0, inputting = 0, nvbombs = 0;

	// move objects
	obi = (6+MAXBOMBS*MAXKEYS)*2;
	oki = 6*2;
	dki = 0*MAXKEYS;
	for(i = 0; i < MAXBOMBS; ++i) {
		if(bomby[i] != -1) {
			if(!pause) { bomby[i] += difficulty.fallspeed; }
			if(204 * 16 <= bomby[i]) {
				if(diskeys[dki] != 0) {
					kaboomx = bombx[i] - 16;
					scrscli = 0;
					say(SOUND_KABOOM, PLAY);
				}
				bomby[i] = -1;
				++upd;
			} else {
				u16 ds = diskeys[dki] >> 4, di = diskeys[dki] & 15;
				// reusing oambuf.x is faster than calc x
				//u16 x = bombx[i] + 8 - ds * 8 / 2;
				oambuf[obi] = ((bomby[i] / 16) << 8) | bombx[i];
				for(j = 0; j < ds; ++j) {
					oambuf[oki] = ((bomby[i] / 16 - 8) << 8) | (oambuf[oki] & 0x00FF);//x;
					//x += 8;
					oki += 2;
				}
				if(2 <= di) {
					++inputting;
				}
				if(ds != 0) {
					++nvbombs;
				}
			}
			obi += 2;
		}
		dki += MAXKEYS;
	}

	return upd ? -1 : ((inputting << 8) | nvbombs);
}

void wave_think() {
	u16 i, j, upd = 0, inputting = 0;

	if(kaboomx != 0x8000) {
		upd = shake();
	} else {
		u16 nvbombs = fall();
		if(nvbombs == -1) {
			nvbombs = 0;
			++upd;
		}
		if(0x100 <= nvbombs) {
			inputting = 1;
			nvbombs &= 0x00FF;
		}

		// spawn bomb
		if(!pause && kaboomx == 0x8000 && nvbombs < difficulty.maxpars) {
			if((nvbombs == 0 && 60 < nospawnperiod) || rand() < difficulty.spawnrate) {
				if(difficulty.bombs) {
					for(i = 0; i < MAXBOMBS; ++i) {
						if(bomby[i] != -1) {
							continue;
						}
						bombx[i] = rand() % (256-8*MAXKEYS) + 8*MAXKEYS/2;
						bomby[i] = 24 * 16;
						j = difficulty.minkeys;
						if(difficulty.maxkeys != difficulty.minkeys) {
							j += rand() % (difficulty.maxkeys - difficulty.minkeys + 1); // divmod is slower, but this is rarely happen event. ok.
						}
						diskeys[i*MAXKEYS+0] = (j << 4) | (!inputting);
						for(++j; j; --j) {
							diskeys[i*MAXKEYS+j] = difficulty.keycands[rand() % difficulty.nkeycands]; // divmod... ditto
						}
						updobj();
						nospawnperiod = 0;
						--difficulty.bombs;
						say(SOUND_FALL, PLAY);
						++upd;
						break;
					}
				} else if(nvbombs == 0 && waitcnt == 0) {
					++waitcnt;
					say(SOUND_SUCCESS, PLAY);
				}
			}
		}
		if(nvbombs == 0) {
			++nospawnperiod;
		}
		if(waitcnt && 90 < ++waitcnt) {
			++waveno;
			prewave_init();
		}
	}

	if(upd) {
		updobj();
	}
}

void wave_input() {
	u16 i, j;

	btn = JOY1;
	randseed ^= btn;
	dbtn = (pbtn ^ btn) & btn;
	pbtn = btn;

	if(kaboomx != 0x8000) {
		return;
	}

	if(dbtn & BTN_START) {
		pause ^= 1;
		updobj();

		// FIXME should do at wave_ppu, this is quickhack...
		VMADD = 0x4400 + 32*13 + 14;
		INIDISP = 0x80; // forceblank
		if(pause) {
			VMDATA = 0x2400 | 'P';
			VMDATA = 0x2400 | 'a';
			VMDATA = 0x2400 | 'u';
			VMDATA = 0x2400 | 's';
			VMDATA = 0x2400 | 'e';
		} else {
			for(i = 0; i < 5; ++i) {
				VMDATA = 0x2400 | ' ';
			}
		}
#ifdef DEBUG
		INIDISP = 0x0d; // noblank bright=14 (darker for debug)
#else
		INIDISP = 0x0F; // noblank bright=16
#endif
	} else if(!pause && (dbtn & ~(BTN_START | BTN_SELECT))) {
		u16 dki, nbomb = 0, ninv = 0, complete = 0, snd = -1;
		for(dki = 0, i = 0; i < MAXBOMBS; dki += MAXKEYS, ++i) {
			if(bomby[i] == -1) {
				continue; // not born
			}
			if(diskeys[dki] == 0) {
				continue; // already done
			}
			++nbomb;
			{
				u16 di = diskeys[dki] & 15;
				if(di == 0) {
					// invalidated
					++ninv;
					continue;
				}
				if(!((0x8000U >> diskeys[dki+di]) & dbtn))
				{
					// this is not for me
					diskeys[dki] &= 0xF0;
					++ninv;
					continue;
				}
				// gotta!
				{
					u16 ds = diskeys[dki] >> 4;
					++di;
					++diskeys[dki];
					if(ds < di) {
						// complete! disarm this bomb
						diskeys[dki] = 0;
						snd = SOUND_ACCEPT;
						++ninv;
						++complete;
					} else {
						// sound type
						snd = SOUND_TYPE;
					}
				}
			}
		}
		// if all invalidated, buzz and reset
		if(0 < nbomb && nbomb == ninv) {
			if(!complete) {
				snd = SOUND_INVALID;
			}

			for(dki = 0, i = 0; i < MAXBOMBS; dki += MAXKEYS, ++i) {
				if(bomby[i] == -1) {
					continue;
				}
				if(diskeys[dki] == 0) {
					continue;
				}
				diskeys[dki] |= 0x01;
			}
		}
		if(snd != -1) {
			say(snd, PLAY);
		}
		// anyway, some modification has done. refresh objs.
		updobj();
	}
}

void wave_think_input(void) {
	wave_think();

	wave_input();
}

void wave_ppu(void) {
	OAMADD = 0;
	dma0(0x220, oambuf, 0x04, 0x02); // 2104=OAMDATA a2b,direct,a=inc,word-same-byte

	BG1HOFS = scrsclx; BG1HOFS = scrsclx>>8;
	BG1VOFS = scrscly; BG1VOFS = scrscly>>8;
	BG2HOFS = scrsclx; BG2HOFS = scrsclx>>8;
	BG2VOFS = scrscly; BG2VOFS = scrscly>>8;
}

void wave_init(void) {
	u16 i;

	scene_think_input = wave_think_input;
	scene_ppu = wave_ppu;

	INIDISP = 0x8F; // forceblank bright=16

	for(i = 0; i < MAXBOMBS; ++i) {
		bomby[i] = -1;
	}

	pause = 0;
	scrsclx = 0;
	scrscly = 0;
	kaboomx = 0x8000;
	nospawnperiod = 90; // make instantaneous spawn
	waitcnt = 0;
	i = waveno - 1;
	if(ARRAY_SIZE(difficulties) <= i) {
		i = ARRAY_SIZE(difficulties) - 1;
	}
	difficulty = difficulties[i];

	VMADD = 0x4000;
	for(i = 0; i < 32*20; ++i) {
		VMDATA = 0x0400;
	}
	for(; i < 32*21; ++i) {
		VMDATA = 0x049C;
	}
	for(; i < 32*22; ++i) {
		VMDATA = 0x049D;
	}
	for(; i < 32*23; ++i) {
		VMDATA = 0x049E;
	}
	for(; i < 32*24; ++i) {
		VMDATA = 0x049F;
	}
	for(; i < 32*26; ++i) {
		VMDATA = 0x04BC;
	}
	for(; i < 32*30; ++i) { // +2 extra
		VMDATA = 0x04BD;
	}
	for(; i < 32*32; ++i) { // all 2 extra
		VMDATA = 0x0400;
	}

	// erase "WAVE xx" on BG1
	VMADD = 0x4400 + 32*13 + 13;
	for(i = 0; i < 7; ++i) {
		VMDATA = 0x2400 | ' ';
	}

	CGADD = 0x00; // backdrop
	COLOR1(0,0,12); // to darkblue
}

void prewave_think_input(void) {
	if((kaboomx == 0x8000 ? 60 : 120) < ++waitcnt) {
		wave_init();
		wave_think_input();
	}
}

void prewave_ppu(void) {
	/* nothing to do... */
}

void prwaveno(u16 attr) {
	if(waveno < 10) {
		VMDATA = attr | ('0' + waveno);
	} else {
		// FIXME ugh... inline itoa
		u16 divisor = 10 << 3, tenbit = 1 << 3, tenth = 0, teeny = waveno;
		if(99 < teeny) { teeny = 99; }
		while(9 < teeny && tenbit) {
			if(divisor <= teeny) {
				teeny -= divisor;
				tenth |= tenbit;
			}
			divisor >>= 1, tenbit >>= 1;
		}
		VMDATA = attr | ('0' + tenth);
		VMDATA = attr | ('0' + teeny);
	}
}

void prewave_init(void) {
	u16 i;

	scene_think_input = prewave_think_input;
	scene_ppu = prewave_ppu;

	INIDISP = 0x8F; // forceblank bright=16

	BG1HOFS = 0x00; BG1HOFS = 0x00>>8; // BG1scrx=0
	BG1VOFS = 0x00; BG1VOFS = 0x00>>8; // BG1scry=0
	BG2HOFS = 0x00; BG2HOFS = 0x00>>8; // BG2scrx=0
	BG2VOFS = 0x00; BG2VOFS = 0x00>>8; // BG2scry=0

	// clear bg0 for make black
	VMADD = 0x4400;
	clrbg0();

	CGADD = 0x00; // backdrop
	COLOR1(0,0,0); // to black

	// clear with status area
	VMADD = 0x4400;
	for(i = 0; i < 32*4; ++i) {
		VMDATA = 0x2001;
	}
	for(; i < 32*30; ++i) { // +2 extra
		VMDATA = 0x2000;
	}
	for(; i < 32*32; ++i) { // all 2 extra
		VMDATA = 0x2001;
	}

	// life
	VMADD = 0x4400 + 32*2 + 2;
	for(i = 0; i < life; ++i) {
		VMDATA = 0x2002;
	}

	VMADD = 0x4400 + 32*13 + 13;
	VMDATA = 0x2400 | 'W';
	VMDATA = 0x2400 | 'A';
	VMDATA = 0x2400 | 'V';
	VMDATA = 0x2400 | 'E';
	VMDATA = 0x2400 | ' ';
	prwaveno(0x2400);

	clrobj();

	waitcnt = 0;

	if(kaboomx == 0x8000) {
		say(SOUND_GSTART, PLAY);
	} else {
		say(SOUND_SSTART, ADSR(15, 0, 7, 14));
	}
}

void game_init(void) {
	INIDISP = 0x8F; // forceblank bright=16

	BGMODE = 0x00; // BG mode=0 BG1-4=8x8
	MOSAIC = 0x00; // no mosaic
	BG1SC = 0x40; // BG1mapbase=16KW scr=32x32
	BG2SC = 0x44; // BG2mapbase=17KW scr=32x32
	BG12NBA = 0x00; // BG1chrbase=0KW BG2chrbase=0KW

	OBSEL = 0x00; // OBJ size=8x8,16x16 gap=0*4KW chrbase=0*8KW

	SETINI = 0x00; // noextsync noextbg noxhires 224lines lowresobj nointl
	TM = 0x13; // main: OBJ,BG2,BG1
	TS = 0x00; // sub: nothing

	VMAIN = 0x80; // vramincr=1 on high,notrans

	clrbg0();
	loadpal();
	clrobj();

	life = 3;
	//waveno = 1;

	kaboomx = 0x8000; // reset the "last wave has failed" flag

	prewave_init();
}

void desc_think_input(void) {
	++waitcnt;
	if(30 < waitcnt) {
		waitcnt = 0;
		oambuf[128*2+125/8] = (oambuf[128*2+125/8] & ~0x5400) | 0xA800;
	} else if(15 == waitcnt) {
		oambuf[128*2+125/8] = (oambuf[128*2+125/8] & ~0x5400) | 0xFC00;
	}

	++descrl;
	if(160 <= descrl) {
		descrl = 0;
	}

	// input
	btn = JOY1;
	randseed ^= btn;
	dbtn = (pbtn ^ btn) & btn;
	pbtn = btn;

	if(secreti == 14 && (dbtn & BTN_UP)) {
		++waveno;
	} else if(secreti == 14 && (dbtn & BTN_DOWN)) {
		--waveno;
		if(waveno < 1) {
			waveno = 1;
		}
	} else if(dbtn) {
		game_init();
	}
}

void desc_ppu(void) {
	u16 x, y;

	OAMADD = 0;
	dma0(0x220, oambuf, 0x04, 0x02); // 2104=OAMDATA a2b,direct,a=inc,word-same-byte

	x = -descrl / 4;
	y = descrl * 2 / 4;
	BG2HOFS = x; BG2HOFS = x >> 8;
	BG2VOFS = y; BG2VOFS = y >> 8;

	if(secreti == 14) {
		VMADD = 0x4000 + 32*23 + 23;
		prwaveno(0x0000);
	}
}

void desc_init(void) {
	u16 i, j, k;

	scene_think_input = desc_think_input;
	scene_ppu = desc_ppu;

	INIDISP = 0x8F; // forceblank bright=16

	HDMAEN = 0x00; // disable title screen trick

	BGMODE = 0x00; // BG mode=0 BG1-4=8x8
	MOSAIC = 0x00; // no mosaic
	BG1SC = 0x40; // BG1mapbase=16KW scr=32x32
	BG2SC = 0x47; // BG2mapbase=17KW scr=64x64
	BG12NBA = 0x00; // BG1chrbase=0KW BG2chrbase=0KW
	TM = 0x13; // main=OBJ,BG2,BG1
	BG1VOFS = 0x00; BG1VOFS = 0x00>>8; // BG1scry=0

	loadpal(); // for OBJ: reusing regular
	CGADD = 0x00; // backdrop,BG0.0
	COLOR4(31,26,16, 31,31,31, 0,0,0, 28,8,8); // cream(backdrop),white,black,lightred
	CGADD = 0x20; // backdrop,BG1.0
	COLOR4(31,26,16, 31,29,29, 0,0,0, 0,0,0); // cream(transp...),lightcream
	// reusing OBJ pal but custom OBJ a bit for bg=white...
	CGADD = 0x90; // cgramaddr=0x90 (OBJ1)
	COLOR4(0,0,0, 20,20,20, 0,0,0, 0,0,0); // done-btns
	CGADD = 0xA0; // cgramaddr=0xA0 (OBJ2)
	COLOR4(0,0,0, 31,20,20, 0,0,0, 0,0,0); // now-btns
	CGADD = 0xB0; // cgramaddr=0xB0 (OBJ3)
	COLOR4(0,0,0, 0,0,0, 0,0,0, 0,0,0); // will-btns

	// bg1: paper
	clrbg0();
	for(k = 0, i = 3; i < 25; ++i) {
		VMADD = 0x4000 + i*32 + 3;
		for(j = 0; j < 26; ++j) {
			VMDATA = 0x0180 + descnam[k];
			++k;
		}
	}

	// bg2: moving bg
	{
		i = 0;
		for(;;) {
			VMADD = (i < 32 ? 0x4800 : 0x5000) + 32*(i%32) + 27;
			for(j = 0;;) {
				VMDATA = 0x0000;
				VMDATA = 0x0000;
				VMDATA = 0x0084;
				VMDATA = 0x0085;
				VMDATA = 0x0086;
				if(j == 0) { VMADD = (i < 32 ? 0x4400 : 0x4C00) + 32*(i%32); }
				else if(j == 7) { break; }
				++j;
			}
			++i;
			VMADD = (i < 32 ? 0x4800 : 0x5000) + 32*(i%32) + 27;
			for(j = 0;;) {
				VMDATA = 0x0000;
				VMDATA = 0x0087;
				VMDATA = 0x00A4;
				VMDATA = 0x00A5;
				VMDATA = 0x0000;
				if(j == 0) { VMADD = (i < 32 ? 0x4400 : 0x4C00) + 32*(i%32); }
				else if(j == 7) { break; }
				++j;
			}
			++i;
			VMADD = (i < 32 ? 0x4800 : 0x5000) + 32*(i%32) + 27;
			for(j = 0;;) {
				VMDATA = 0x00A6;
				VMDATA = 0x00A7;
				VMDATA = 0x015E;
				VMDATA = 0x0000;
				VMDATA = 0x0000;
				if(j == 0) { VMADD = (i < 32 ? 0x4400 : 0x4C00) + 32*(i%32); }
				else if(j == 7) { break; }
				++j;
			}
			++i;
			VMADD = (i < 32 ? 0x4800 : 0x5000) + 32*(i%32) + 27;
			for(j = 0;;) {
				VMDATA = 0x015F;
				VMDATA = 0x017E;
				VMDATA = 0x0000;
				VMDATA = 0x0000;
				VMDATA = 0x0000;
				if(j == 0) { VMADD = (i < 32 ? 0x4400 : 0x4C00) + 32*(i%32); }
				else if(j == 7) { break; }
				++j;
			}
			++i;
			VMADD = (i < 32 ? 0x4800 : 0x5000) + 32*(i%32) + 27;
			for(j = 0; j < 5; ++j) {
				VMDATA = 0x0000;
			}
			VMADD = (i < 32 ? 0x4400 : 0x4C00) + 32*(i%32);
			for(j = 0; j < 32; ++j) {
				VMDATA = 0x0000;
			}
			++i;
			if(35 < i) { break; }
		}
	}

	// reusing regular updobj for desc...
	kaboomx = 0x8000;
	bombx[0] = 50;
	bomby[0] = 120 * 16;
	diskeys[0] = 0x42;
	diskeys[1] = BI_A;
	diskeys[2] = BI_X;
	diskeys[3] = BI_PU;
	diskeys[4] = BI_R;
	for(i = 1; i < 9; ++i) {
		bomby[i] = -1;
	}
	updobj();

	// kaboom is manually placed
	oambuf[0*2+0] = ((154 +  0) << 8) | (172 +  0); oambuf[0*2+1] = 0x3860;
	oambuf[1*2+0] = ((154 +  0) << 8) | (172 + 16); oambuf[1*2+1] = 0x3862;
	oambuf[2*2+0] = ((154 +  0) << 8) | (172 + 32); oambuf[2*2+1] = 0x3864;
	oambuf[3*2+0] = ((154 + 16) << 8) | (172 +  0); oambuf[3*2+1] = 0x3866;
	oambuf[4*2+0] = ((154 + 16) << 8) | (172 + 16); oambuf[4*2+1] = 0x3868;
	oambuf[5*2+0] = ((154 + 16) << 8) | (172 + 32); oambuf[5*2+1] = 0x386A;
	oambuf[128*2+0/8] = (oambuf[128*2+0/8] & ~0x0555) | 0x0AAA;

	// press (use palette for bombs, reuse.)
	oambuf[125*2+0] = (184 << 8) | (104 + 0);
	oambuf[125*2+1] = 0x306C;
	oambuf[126*2+0] = (184 << 8) | (104 + 16);
	oambuf[126*2+1] = 0x306E;
	oambuf[127*2+0] = (184 << 8) | (104 + 32);
	oambuf[127*2+1] = 0x308C;
	oambuf[128*2+125/8] = (oambuf[128*2+125/8] & ~0x5400) | 0xA800;

	waitcnt = 0;
	descrl = 0;

	waveno = 1;

	say(SOUND_OBJCTV, PLAY);
}

void title_think_input(void) {
	u16 i;

	if(0x1000 <= waitcnt) {
		if(0x1000 + 32 <= waitcnt) {
			desc_init();
			return;
		}
	} else if(waitcnt <= 90) {
		u16 n = (90 - waitcnt), y;
		if(0x8000 <= n) {
			n = 0;
		}

		// title scroll down
		hdmabuf[1] = 12*8 * n / 90;
		// subti scroll left
		for(i = 0; i < 7; ++i) {
			u16 x = 48 + (256 - 48) * n / 90 + 24*i;
			oambuf[i*2+0] = ((15*8) << 8) | (x & 255);
			oambuf[128*2 + i/8] = (oambuf[128*2 + i/8] & ~(1 << (i % 8 * 2))) | ((256 <= x) << (i % 8 * 2));
		}
		// credit
		y = 200 + n;
		if(240 < y) {
			y = 240;
		}
		for(i = 0; i < 5; ++i) {
			oambuf[(10+i)*2+0] = (y << 8) | (88 + 16*i);
		}

	} else {
		if(90 + 30 < waitcnt) {
			waitcnt = 90;
		}
		if(waitcnt < 90 + 15) {
			oambuf[128*2 + 7/8] = (oambuf[128*2 + 7/8] & ~(0x0001 << (7 % 8 * 2))) | (0x0000 << (7 % 8 * 2));
			oambuf[128*2 + 8/8] = (oambuf[128*2 + 8/8] & ~(0x0005 << (8 % 8 * 2))) | (0x0000 << (8 % 8 * 2));
		} else {
			oambuf[128*2 + 7/8] = (oambuf[128*2 + 7/8] & ~(0x0001 << (7 % 8 * 2))) | (0x0001 << (7 % 8 * 2));
			oambuf[128*2 + 8/8] = (oambuf[128*2 + 8/8] & ~(0x0005 << (8 % 8 * 2))) | (0x0005 << (8 % 8 * 2));
		}
	}
	// step always...
	++waitcnt;

	// input
	{
		u16 cbtn, ubtn, anykey;

		btn = JOY1;
		randseed ^= btn;
		cbtn = pbtn ^ btn;
		dbtn = cbtn & btn;
		ubtn = cbtn & pbtn;
		pbtn = btn;

		anykey = (dbtn & ~(BTN_L | BTN_R)) | (ubtn & (BTN_L | BTN_R));
		if((btn & (BTN_L | BTN_R)) || 13 <= secreti) {
			if(dbtn || (ubtn & (BTN_L | BTN_R))) {
				u16 b = 0;
				switch(secreti) {
					case 0: b = (dbtn & BTN_R) && !(btn & BTN_L); break;
					case 1: case 2: b = (dbtn & BTN_UP) && (btn & BTN_R) && !(btn & BTN_L); break;
					case 3: case 4: b = (dbtn & BTN_DOWN) && (btn & BTN_R) && !(btn & BTN_L); break;
					case 5: case 7: b = (dbtn & BTN_LEFT) && (btn & BTN_R) && !(btn & BTN_L); break;
					case 6: case 8: b = (dbtn & BTN_RIGHT) && (btn & BTN_R) && !(btn & BTN_L); break;
					case 9: b = (dbtn & BTN_L) && (btn & BTN_R); break;
					case 10: b = (ubtn & BTN_R) && (btn & BTN_L); break;
					case 11: b = (dbtn & BTN_B) && (btn & BTN_L) && !(btn & BTN_R); break;
					case 12: b = (dbtn & BTN_A) && (btn & BTN_L) && !(btn & BTN_R); break;
					case 13: b = (ubtn & BTN_L) && !(btn & BTN_R); break;
					case 14: b = 1; break;
				}
				if(b) {
					if(secreti < 14) {
						anykey = 0; // ignore this dbtn/ubtn(L/R)
						++secreti;
					}
					if(secreti == 13) {
						// XXX do this here!?
						u16 zero = 0x00;
						u8 *p;

						INIDISP = 0x8F; // forceblank bright=16

						VMADD = 0x4800; // vramaddr=18KW
						dma0(2 * 32 * 32, &zero, 0x18, 0x09); // (u16*)0x2118=vramio, a2b,direct,a=fixed,word

						VMADD = 0x4800 + 32*13 + 3;
						for(p = buildstr; *p; ++p) {
							VMDATA = 0x1400 | *p;
						}

						// XXX overlapping BG01.pal1!
						CGADD = 0x14;
						COLOR4(0,0,0, 0,0,0, 0,0,0, 31,31,31); // (transp),(unused),(unused),white

						BG34NBA = 0x00; // BG3chrbase=0KW (BG4chrbase=0KW)
						BG3SC = 0x48; // BG1mapbase=18KW scr=32x32
						BG3HOFS = 0x00; BG3HOFS = 0x00>>8; // BG3scrx=0
						BG3VOFS = 0x00; BG3VOFS = 0x00>>8; // BG3scry=0
						TM = 0x15; // main: OBJ,BG3,BG1

						INIDISP = 0x0F; // noblank bright=16
					}
				} else {
					secreti = 0;
				}
			}
		} else {
			secreti = 0;
		}
		if(anykey) {
			if(waitcnt < 90) {
				waitcnt = 90;
			} else if(waitcnt < 0x1000) {
				waitcnt = 0x1000;
			}
		}
	}
}

void title_ppu(void) {
	OAMADD = 0;
	dma0(0x220, oambuf, 0x04, 0x02); // 2104=OAMDATA a2b,direct,a=inc,word-same-byte

	if(0x1000 <= waitcnt) {
		MOSAIC = (((waitcnt - 0x8000) / 2) << 4) | 0x01;
	}
}

void title_init(void) {
	u16 i, j, k;

	scene_think_input = title_think_input;
	scene_ppu = title_ppu;

	INIDISP = 0x8F; // forceblank bright=16

	BGMODE = 0x01; // BG mode=1 BG1-4=8x8
	BG1SC = 0x42; // BG1mapbase=16KW scr=32x64
	TM = 0x11; // main: OBJ,BG0

	// palette
	CGADD = 0x00; // backdrop,bg0
	COLOR4(6,6,25, 6,5,25, 7,5,24, 10,5,22);
	COLOR4(11,5,21, 13,5,19, 15,5,18, 16,5,17);
	COLOR4(19,5,15, 21,5,14, 22,5,12, 24,5,11);
	COLOR4(26,5,9, 27,5,9, 28,5,8, 29,5,8);
	CGADD = 0x10; // bg0.1
	COLOR4(6,6,25, 9,6,25, 12,6,25, 15,6,25);

	CGADD = 0x80; // obj0
	COLOR4(0,0,0, 31,18,0, 0,0,0, 0,0,0); // (transp),orange
	CGADD = 0x90; // obj1
	COLOR4(0,0,0, 31,31,31, 0,0,0, 0,0,0); // (transp),white

	// title bg
	VMADD = 0x4000; // 16KW
	for(i = 0; i < 32*2; ++i) { // padding
		VMDATA = 0x0140;
	}
	for(k = 0, i = 0; i < 10; ++i) { // ake2019
		VMDATA = 0x0140;
		VMDATA = 0x0140;
		for(j = 0; j < 28; ++j) {
			VMDATA = 0x0140 + titlenam[k];
			++k;
		}
		VMDATA = 0x0140;
		VMDATA = 0x0140;
	}
	for(i = 0; i < 32*16; ++i) { // padding: with scroll buffer
		VMDATA = 0x0140;
	}
	// gradations
	for(; i < 32*17; ++i) {
		VMDATA = 0x048A;
	}
	for(; i < 32*18; ++i) {
		VMDATA = 0x048B;
	}
	for(; i < 32*20; ++i) {
		VMDATA = 0x049A;
	}
	for(; i < 32*21; ++i) {
		VMDATA = 0x049B;
	}
	for(; i < 32*22; ++i) {
		VMDATA = 0x048E;
	}
	for(; i < 32*24; ++i) {
		VMDATA = 0x048F;
	}
	for(; i < 32*25; ++i) {
		VMDATA = 0x049E;
	}
	for(; i < 32*26; ++i) {
		VMDATA = 0x049F;
	}
	// bottom
	for(; i < 32*32; ++i) {
		VMDATA = 0x04AE;
	}

	clrobj();

	// subti
	for(i = 0; i < 7; ++i) {
		oambuf[i*2+1] = 0x30A0 + i * 2;
	}
	// press
	oambuf[7*2+0] = (168 << 8) | (104 + 0);
	oambuf[7*2+1] = 0x326C;
	oambuf[8*2+0] = (168 << 8) | (104 + 16);
	oambuf[8*2+1] = 0x326E;
	oambuf[9*2+0] = (168 << 8) | (104 + 32);
	oambuf[9*2+1] = 0x328C;
	// copyr
	for(i = 0; i < 5; ++i) {
		oambuf[(10+i)*2+0] = (240 << 8) | (88 + 16*i);
		oambuf[(10+i)*2+1] = 0x3280 + i*2;
	}

	oambuf[128*2 + 0/8] = 0xFFFF; // 0-7
	oambuf[128*2 + 8/8] = 0x6AAF; // 8-15

	// setup title scroller
	hdmabuf[0] = 14*8;
	hdmabuf[1] = 12*8;
	hdmabuf[2] = (12*8)>>8;
	hdmabuf[3] = 1;
	hdmabuf[4] = 14*8;
	hdmabuf[5] = (14*8)>>8;
	hdmabuf[6] = 0;

	DMAP1 = 0x02; // a2b,direct,(inc),xfer2-same
	BBAD1 = 0x0E; // bg1vofs
	A1T1 = (u16)hdmabuf;
	{
		void *p = hdmabuf;
		A1B1 = ((u8*)&p)[2];
	}
	//DAS1 = ; DASB1 = ;
	HDMAEN = 0x02;

	waitcnt = 0;
	subtix = 256;
	subtidx = 0;
	secreti = 0;
}

void anniv_think_input(void) {
	// think
	++waitcnt;
	if(waitcnt == 30) {
		// enable sound only if APU does working and L+R is NOT pressed about 100ms in wait period.
		sound = (isapuinited() && sound < 6);
		if(sound) {
			soundinit();
		}
		say(SOUND_15TH, PLAY);
		// FIXME oops modifying PPU in think!! (changing TM may not require forceblank, but it's easy to know sound load time if DEBUG)
		INIDISP = 0x8F; // forceblank bright=16
		TM = 0x01; // main: BG0
#ifdef DEBUG
		INIDISP = 0x0d; // noblank bright=14 (darker for debug)
#else
		INIDISP = 0x0F; // noblank bright=16
#endif

		// load rest
		if(sound) {
			apuexec(0xFF80);
			triload(sizeof_sounds_bin - sizeof_15th, BRRAT + sizeof_15th, sounds_bin + sizeof_15th);
			triexec(0xFFC9);
			apuwaitinit();
		}
	} else if(waitcnt == (sound ? 31 : 90)) { // note: sound load time becomes logo wait time
		// load descchr: should be at title_init but required only once, so do this here.
		VMADD = 0x0C00; // vramaddr=3KW
		dma0(sizeof_descchr, descchr, 0x18, 0x01); // (u16*)0x2118=vramio, mem2io,direct,a=inc,word

		title_init();
		return;
	}

	// input
	btn = JOY1;
	randseed ^= btn;
	dbtn = (pbtn ^ btn) & btn;
	pbtn = btn;

	if(waitcnt < 30 && (btn & (BTN_L | BTN_R)) == (BTN_L | BTN_R)) {
		++sound;
	}
}

void anniv_ppu(void) {
	/* nothing to do... */
}

void anniv_init(void) {
	u16 i, j, k;

	scene_think_input = anniv_think_input;
	scene_ppu = anniv_ppu;

	BGMODE = 0x01; // BG mode=1 BG1-4=8x8
	BG1SC = 0x40; // BG1mapbase=16KW scr=32x32
	TM = 0x00; // main: nothing (later BG0 shows)

	loadchr();
	clrobj();

	// grayscale with black backdrop
	CGADD = 0x00; // cgramaddr=0x00 (backdrop/BG0)
	COLOR1(0,0,0);
	for(i = 3; i <= 31; i += 2) {
		COLOR1(i,i,i);
	}

	// clear bg0 by anniv's first chr
	VMADD = 0x4000;
	for(i = 0; i < 32*32; ++i) {
		VMDATA = 0xC0;
	}
	// load anniversary map
	for(k = 0xC0, i = 8+0; i < 8+12; ++i) {
		VMADD = 0x4000 + 32*i + 11;
		for(j = 0; j < 10; ++j) {
			VMDATA = k;
			++k;
		}
	}

	sound = 0;
	waitcnt = 0;

	INIDISP = 0x0F; // noblank bright=16
}

void main(void) {
	init();

	randseed = 1;
	pbtn = 0;

	anniv_init();

	for(;;) {
		scene_think_input();

		rand(); // make noise

#ifdef DEBUG
		INIDISP = 0x0F; // noblank bright=16 (resetting debug dark)
#endif

		vblanked = 0;
		while(!vblanked) /*nothing*/ ;

		INIDISP = 0x8F; // forceblank bright=16

		// PPU
		scene_ppu();

#ifdef DEBUG
		INIDISP = 0x0d; // noblank bright=14 (darker for debug)
#else
		INIDISP = 0x0F; // noblank bright=16
#endif
	}
}
