# coding: utf-8

def sim a, b
	a.zip(b).map{|x,y|(x-y).abs}.sum
end

# input: TIFF 4bpp output by FLAX
def loadtiff fn
	File.open(fn, "rb") { |f|
		bmark, ver, ifd = f.read(8).unpack("a2vV")
		(bmark == "II" and ver == 0x2A) or raise "?magic"
		f.pos = ifd
		ntags = f.read(2).unpack("v")[0]
		# XXX what a quick hack!
		tags = ntags.times.map{ h=%i(tag type nels data).zip(f.read(12).unpack("vvVV")).to_h; [h[:tag], h[:data]] }.to_h
		nifd = f.read(4).unpack("V")[0]
		nifd == 0 or raise "?not-supported-multi-page-tiff"
		width = tags[0x0100]
		width % 8 == 0 or raise "?width-mod-8 #{width%8}"
		height = tags[0x0101]
		height % 8 == 0 or raise "?height-mod-8 #{height%8}"
		tags[0x0102] == 4 or raise "?not-supported-non-4bpp #{tags[0x0102].inspect}"
		tags[0x0103] == 1 or raise "?not-supported-compressed #{tags[0x0103].inspect}"
		f.pos = tags[0x0111]
		bits = f.read(width * height / 2)
		return [width, height, bits]
	}
end

thresh = (ARGV[1] || 0).to_i
fnbase = ARGV[0].sub(/\.[^.]+/, "")
width, height, bits = loadtiff ARGV[0]
rawchrs = bits.bytes.flat_map{|e| [e>>4,e&15] }.each_slice(8).each_slice(width/8).each_slice(8).flat_map{|b| b.transpose }.map{|e|e.flatten}

chr = []
nam = []
c2n = {}
sis = []

rawchrs.each { |c|
	# i = (c2n[c] ||= chr.size)
	if not c2n[c]
		if chr.empty?
			i = (c2n[c] ||= chr.size)
			sis = [[1.0/0, nil]]
		else
			s = chr.each_with_index.map{|e,i| [sim(e, c), i] }.sort.first
			if s[0] <= thresh
				i = s[1]
			else
				sis << s
				i = (c2n[c] ||= chr.size)
			end
		end
	else
		i = c2n[c]
	end
	nam << i
	chr[i] = c
}

ncolors = chr.map{|c| c.max }.max + 1

bchr = "".b
chr.each { |c| # c = [4b * 8 * 8]
	if ncolors <= 4
		fourplanes = [c]
	else
		fourplanes = c.map{|e| [e&3, e>>2] }.to_a.transpose # [[l2b * 8 * 8], [u2b * 8 * 8]]
	end
	fourplanes.each { |fplane|
		fplane.each_slice(8) { |fpline| # [2b * 8]
			oneplanes = fpline.map{|e| [e&1, e>>1] }.to_a.transpose # [[l1b * 8], [u1b * 8]]
			oneplanes.each { |oplane| # [1b * 8]
				bchr << [oplane.inject(0){|r,i| (r << 1) | i }].pack("C")
			}
		}
	}
}

File.binwrite("#{fnbase}.nam", nam.each_slice(width/8).flat_map{|x| x + [0]*(32-x.size) * 0 }.pack("C*"))
File.binwrite("#{fnbase}.chr", bchr)

ARGV[1] or p sis.sort
